"""eulerworkbench
All workbench are build on top of this module.

Copyright (C) 2014  Stijn Wouters  <w.stijn@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Last modified: 10 August 2014.
By: Stijn Wouters.
"""
import os
import sys
import imp
import pkgutil
import urllib.request
try:
    from time import perf_counter
except ImportError:
    # provide backwards compatibility
    from time import clock as perf_counter

import bs4


try:
    raise FileNotFoundError('')
except NameError:
    # Provide backward compatibility for Python3.2, which doesn't support
    # FileNotFoundError standard exception.
    FileNotFoundError = OSError
except OSError:
    # silently ignore this exception
    pass


# version info, put here to avoid circular dependencies
__version__ = (0, 3, 0)


class EulerWorkbench:
    """Base workbench providing decorators and basic tools for your Project
    Euler.
    """

    SOLUTIONS = []
    """Shared database containing all your solutions on Project Euler (as found
    so far).
    """

    def __init__(self):
        """Constructor. Initialize default settings.
        """
        # alias to static data member, just for easy typing
        self.solutions = EulerWorkbench.SOLUTIONS
        return

    @staticmethod
    def solution(solution):
        """Decorator marking the function as a solution for a problem that
        was posed on Project Euler.

        Requires the solution to be callable (otherwise `AttributeError` is
        raised).
        """
        if callable(solution):
            EulerWorkbench.SOLUTIONS.append(solution)
        else:
            raise AttributeError('{} is not callable'.format(solution))

    def discover(self, path=os.getcwd()):
        """Discover Project Euler solutions.

        Notice that if it fails to discovera solution from a subdirectory,
        you'll probably have to add a `__init__.py`, so the subdirectory
        becomes essentially a package. Solution discovery only works for
        packages and modules.
        """
        # Convert string to iterables, so you can still use this
        # implementation, instead of jumpint to somewhere else.
        if isinstance(path, str):
            path = [path]

        for pathname in path:
            if not os.path.exists(pathname):
                msg = 'File or directory not found: {}'.format(pathname)
                raise FileNotFoundError(msg)
            elif os.path.isfile(pathname) and pathname.endswith('.py'):
                # get module name
                module, _ = os.path.splitext(os.path.basename(pathname))

                # skip if module already loaded
                if module in sys.modules:
                    continue

                # load source
                with open(pathname) as filehandler:
                    imp.load_module(module, filehandler, pathname,
                                    ('.py', 'r', imp.PY_SOURCE))
            elif os.path.isdir(pathname):
                # iterate over modules and packages on the given path
                iter_modules = pkgutil.iter_modules(path=[pathname])

                for finder, module, is_package in iter_modules:
                    if is_package:
                        package_path = os.path.join(finder.path, module)

                        self.discover(package_path)
                    else:
                        # get path to module
                        module_path = os.path.join(finder.path, module + '.py')

                        self.discover(module_path)
            else:
                # The given pathname do exist, but is not a directory or a
                # regular file (this could be a symbolic link or a broken file
                # for exmple).
                continue

    def benchmarks(self):
        """Benchmarks the solutions.

        Yields a tuple of (<qualified function name>, answer, benchmark), where
        the benchmark is a return value of `time.perf_counter` for Python
        versions 3.3 and higher:

            Returns the value (in fractional seconds) of a performance counter,
            i.e. a clock with the highest available resolution to measure a
            short duration. It does include time elapsed during sleep and is
            system-wide. The reference point of the returned value is
            undefined, so that only the difference between the result of
            consecutive calls is valid.

        Older versions of Python3 actually use `time.clock`:

            On Unix, return the current processor time as a floating point
            number expressed in seconds. The precision, and in fact the very
            definition of the meaning of 'processor time', depends on that of
            the C function of the same name, but in any case, this is the
            function to use for benchmarking Python or timing algorithms.

            On Windows, this function returns wall-clock seconds elapsed since
            the first call to this function, as a floating point number, based
            on the Win32 function QueryPerformanceCounter(). The resolution is
            typically better than one microsecond.

        Note that, when the memory is exhausted (raising `MemoryError`, the
        answer will be set to `None` and the benchmark to `NaN`.

        The name of the solution follows the conventional Python syntax:

            module:[class.]function
        """
        for solution in self.solutions:
            name = solution.__module__ + ':'
            try:
                name += solution.__qualname__
            except AttributeError:
                name += solution.__name__

            try:
                start = perf_counter()
                answer = solution()
                end = perf_counter()

                yield (name, answer, end - start)
            except MemoryError:
                # memory exhausted
                yield (name, None, float('nan'))

    def fetch(self, problem, timeout=10):
        """Fetch the given problem number from the project euler website.

        Returns the posed problem (as a string) if the problem exist, otherwise
        returns an empty string.

        Raises `socket.timeout` if it fails to fetch a problem from the site
        due to timeout.
        """
        url = 'https://www.projecteuler.net/problem={:d}'.format(problem)

        with urllib.request.urlopen(url, timeout=timeout) as handler:
            html = handler.read()

        # convert bytes to string
        html = html.decode()
        crawler = bs4.BeautifulSoup(html)

        the_problem = crawler.find(role='problem')
        return the_problem.get_text() if the_problem is not None else ''


# alias for less typing
solution = EulerWorkbench.solution
